'' =================================================================================================
''
''   File....... jm_txserial.spin
''   Purpose.... True mode serial transmit driver -- with buffer
''   Author..... Jon "JonnyMac" McPhalen (aka Jon Williams)
''               Copyright (c) 2009 Jon McPhalen
''               -- some elements "borrowed" from fullduplexserial and simple_serial
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 04 JUL 2009
''
'' =================================================================================================


var

  long  cog

  long  txPin                                                   ' transmit pin
  long  txBitTix                                                ' counter ticks per bit
  long  txHead                                                  ' head pointer (0..63)
  long  txTail                                                  ' tail pointer (0..63)
  long  txPntr                                                  ' address of txBuf[0]    

  byte  txBuf[64]                                               ' transmit buffer 


pub init(txd, baud) : okay

'' Creates true mode transmit UART on pin txd at baud rate

  cleanup

  txPin     := txd
  txBitTix  := clkfreq / baud
  txHead    := 0
  txTail    := 0
  txPntr    := @txBuf
       
  bytefill(@txBuf, 0, 64)                                       ' clear buffer
  okay := cog := cognew(@txserial, @txPin) + 1     


pub cleanup

'' Stops serial TX driver; frees a cog

  if cog
    cogstop(cog~ - 1)


pub tx(c)

'' Move c into transmit buffer if room is available
'' -- will wait if buffer is full

  repeat until (txTail <> ((txHead + 1) & $3F))
  txBuf[txHead] := c
  txHead := (txHead + 1) & $3F


pub str(pntr)

'' Transmit z-string at pntr

  repeat strsize(pntr)
    tx(byte[pntr++])
    

pub dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    tx("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      tx(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      tx("0")
    i /= 10


pub hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    tx(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))


pub bin(value, digits)

'' Print a binary number

  value <<= (32 - digits)
  repeat digits
    tx((value <-= 1) & 1 + "0")       


dat

                        org     0

txserial                mov     tmp1, par                       ' start of structure
                        rdlong  tmp2, tmp1                      ' read tx pin
                        mov     txmask, #1                      ' create tx pin mask
                        shl     txmask, tmp2
                        or      outa, txmask                    ' set pin to idle
                        or      dira, txmask                    ' make pin an output   
                        
                        add     tmp1, #4
                        rdlong  bitticks, tmp1                  ' get bit timing
                        
                        add     tmp1, #4
                        mov     headpntr, tmp1                  ' save pointer addresses
                        
                        add     tmp1, #4
                        mov     tailpntr, tmp1
                        
                        add     tmp1, #4
                        rdlong  bufpntr, tmp1                   ' get buffer address  

waitbuf                 rdlong  tmp1, headpntr                  ' tmp1 = txHead
                        rdlong  tmp2, tailpntr                  ' tmp2 = txTail
                        cmp     tmp1, tmp2              wz      ' equal?
                if_e    jmp     #waitbuf                        ' if yes, keep waiting

                        mov     tmp1, bufpntr                   ' tmp1 := @txbuf[0]
                        add     tmp1, tmp2                      ' tmp1 := @txbuf[txTail]
                        rdbyte  txwork, tmp1                    ' txWork := txbuf[txTail] 

txbyte                  call    #transmit

updatetail              add     tmp2, #1                        ' inc txTail
                        and     tmp2, #$3F                      ' wrap to 0 if necessary
                        wrlong  tmp2, tailpntr                  ' save
                        jmp     #waitbuf


' -----------------
' True Mode TX UART
' -----------------
'
transmit                or      txwork, STOP_BITS               ' set stop bit(s)
                        shl     txwork, #1                      ' add start bit
                        mov     txcount, #11                    ' start + 8 data + 2 stop
                        mov     txtimer, bitticks               ' load bit timing
                        add     txtimer, cnt                    ' sync with system counter

txbit                   shr     txwork, #1              wc      ' move bit0 to C
                        muxc    outa, txmask                    ' output the bit
                        waitcnt txtimer, bitticks               ' let timer expire, reload   
                        djnz    txcount, #txbit                 ' update bit count
                        
transmit_ret            ret
 

' -------------------------------------------------------------------------------------------------

STOP_BITS               long    $FFFF_FF00

tmp1                    res     1
tmp2                    res     1   

txmask                  res     1                               ' mask for tx pin
bitticks                res     1                               ' ticks per bit
headpntr                res     1                               ' pointer to head position
tailpntr                res     1                               ' pointer to tail position
bufpntr                 res     1                               ' pointer to txBuf[0]

txwork                  res     1                               ' byte to transmit
txcount                 res     1                               ' bits to transmit
txtimer                 res     1                               ' tx bit timer 
                                 
                        fit     492
                        

dat

{{

  Copyright (c) 2009 Jon McPhalen (aka Jon Williams)

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}                    